#include "nvfs.h"

void nvfs_set_inode_size(struct nvfs_memory_inode *nmi, uint64_t size)
{
	i_size_write(&nmi->vfs_inode, size);
	nv_store(nmi->nvi->i_size, cpu_to_le64(size));
}

void nvfs_set_inode_blocks(struct nvfs_memory_inode *nmi, uint64_t blocks)
{
	WRITE_ONCE(nmi->vfs_inode.i_blocks, blocks);
	nv_store(nmi->nvi->i_blocks, blocks);
}

void nvfs_zero_page(struct nvfs_superblock *nvs, void *ptr, unsigned prealloc, void *aux)
{
	void *end = ptr + (1UL << nvs->log2_page_size);
	do {
		uint64_t zero = 0;
		nv_store(((uint64_t *)ptr)[0], zero);
		nv_store(((uint64_t *)ptr)[1], zero);
		nv_store(((uint64_t *)ptr)[2], zero);
		nv_store(((uint64_t *)ptr)[3], zero);
		nv_store(((uint64_t *)ptr)[4], zero);
		nv_store(((uint64_t *)ptr)[5], zero);
		nv_store(((uint64_t *)ptr)[6], zero);
		nv_store(((uint64_t *)ptr)[7], zero);
		ptr += 64;
	} while (ptr < end);
}

static void nvfs_adj_block_count(struct nvfs_memory_inode *nmi, long adj, bool quota)
{
	if (likely(!quota)) {
		struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
		adj <<= nvs->log2_page_size - 9;
		nmi->vfs_inode.i_blocks += adj;
	}
	nv_store(nmi->nvi->i_blocks, cpu_to_le64(nmi->vfs_inode.i_blocks));
}

/*
 * Note: this may be called concurrently from page fault handles, so we must use
 * spinlock to protect concurrent modification.
 *
 * This will not race with truncations or hole punching.
 * truncate-vs-nvfs_read/nvfs_write is protected with inode_lock
 * truncate-vs-nvfs_fault is protected with dax_sem
 */
void *nvfs_bmap(struct nvfs_memory_inode *nmi, uint64_t page, unsigned *prealloc, void (*create_page)(struct nvfs_superblock *nvs, void *page, unsigned prealloc, void *aux), void *aux, bool *new)
{
	struct nvfs_superblock *nvs;
	struct nvfs_inode *nvi;
	unsigned long nodes_per_page;
	uint64_t npi;
	unsigned level;
	__le64 *mapped;
	uint64_t mapped_x;
	bool quota;
	unsigned limit;
	unsigned p;
	unsigned p1;
	bool can_prealloc;

	cond_resched();

	if (!prealloc) {
		p1 = 1;
		prealloc = &p1;
	}

	nvs = nmi->vfs_inode.i_sb->s_fs_info;
	nvi = nmi->nvi;
	nodes_per_page = 1UL << (nvs->log2_page_size - 3);
	quota = unlikely((nvs->mount_flags & (NVFS_USR_QUOTA | NVFS_GRP_QUOTA)) != 0);
	can_prealloc = page >= i_size_read(&nmi->vfs_inode) >> nvs->log2_page_size;

	if (new)
		*new = false;

	if (page < NVFS_DIRECT_BLOCKS) {
		mapped = &nvi->i_direct_blocks[page];
		limit = NVFS_DIRECT_BLOCKS - page;
		goto found;
	}
	page -= NVFS_DIRECT_BLOCKS;

	npi = nodes_per_page;
	for (level = 0; level < NVFS_INDIRECT_BLOCKS; level++) {
		if (page < npi || !npi)
			goto do_indirect;
		page -= npi;
		npi <<= nvs->log2_page_size - 3;
	}
	if (!create_page) {
		*prealloc = 1;
		return NULL;
	}
	return ERR_PTR(-EFBIG);

do_indirect:
	mapped = &nvi->i_indirect_blocks[level];
	do {
		uint64_t idx = (page >> (level * (nvs->log2_page_size - 3))) & (nodes_per_page - 1);
		mapped_x = le64_to_cpu(READ_ONCE(*mapped));
		if (unlikely(!mapped_x)) {
			if (!create_page) {
				*prealloc = 1;
				return NULL;
			}

			if (unlikely(quota)) {
				int r = dquot_alloc_space_nodirty(&nmi->vfs_inode, 1U << nvs->log2_page_size);
				if (unlikely(r))
					return ERR_PTR(r);
			}
			mapped_x = nvfs_new_block(nvs, NULL);
			if (unlikely(!mapped_x)) {
				if (unlikely(quota))
					dquot_free_space_nodirty(&nmi->vfs_inode, 1U << nvs->log2_page_size);
				return ERR_PTR(-ENOSPC);
			}
			nvfs_zero_page(nvs, nvs->memory_map + mapped_x, 1, NULL);

			spin_lock(&nmi->vfs_inode.i_lock);
			if (unlikely(le64_to_cpu(*mapped) != 0)) {
				uint64_t to_free = mapped_x;
				mapped_x = le64_to_cpu(*mapped);
				spin_unlock(&nmi->vfs_inode.i_lock);
				nvfs_free_blocks(nvs, to_free, 1);
			} else {
				nvfs_adj_block_count(nmi, 1, quota);
				wmb();
				nv_store(*mapped, cpu_to_le64(mapped_x));
				wmb();
				spin_unlock(&nmi->vfs_inode.i_lock);
			}
		}
		mapped = nvs->memory_map + mapped_x + idx * 8;
		limit = nodes_per_page - idx;
	} while (level--);

found:
	if (*prealloc > limit)
		*prealloc = limit;
	mapped_x = le64_to_cpu(READ_ONCE(*mapped));
	if (mapped_x == cpu_to_le64(0)) {
		if (!create_page) {
			*prealloc = 1;
			return NULL;
		}
		if (!can_prealloc)
			*prealloc = 1;
		mapped_x = nvfs_new_block(nvs, prealloc);
		if (unlikely(!mapped_x))
			return ERR_PTR(-ENOSPC);

		if (unlikely(quota)) {
			int r = dquot_alloc_space_nodirty(&nmi->vfs_inode, *prealloc << nvs->log2_page_size);
			nvfs_free_blocks(nvs, mapped_x, *prealloc);
			if (unlikely(r))
				return ERR_PTR(r);
		}

		create_page(nvs, nvs->memory_map + mapped_x, *prealloc, aux);

		spin_lock(&nmi->vfs_inode.i_lock);
		if (unlikely(le64_to_cpu(*mapped) != 0)) {
			uint64_t to_free = mapped_x;
			mapped_x = le64_to_cpu(*mapped);
			spin_unlock(&nmi->vfs_inode.i_lock);
			if (unlikely(quota))
				dquot_free_space_nodirty(&nmi->vfs_inode, *prealloc << nvs->log2_page_size);
			nvfs_free_blocks(nvs, to_free, *prealloc);
			*prealloc = 1;
		} else {
			nvfs_adj_block_count(nmi, *prealloc, quota);
			wmb();
			for (p = 0; p < *prealloc; p++) {
				nv_store(*(mapped + p), cpu_to_le64(mapped_x + (p << nvs->log2_page_size)));
			}
			wmb();
			if (new)
				*new = true;
			spin_unlock(&nmi->vfs_inode.i_lock);
			return nvs->memory_map + mapped_x;
		}
	}
	for (p = 0; p < *prealloc; p++)
		if (le64_to_cpu(READ_ONCE(*(mapped + p))) != mapped_x + (p << nvs->log2_page_size))
			break;
	*prealloc = p;
	return nvs->memory_map + mapped_x;
}

struct punch_info {
	uint64_t start_page;
	uint64_t end_page;
	bool zero;
	long freed;
};

static void walk_direct(struct nvfs_memory_inode *nmi, __le64 *ptr, unsigned long add, uint64_t idx, struct punch_info *punch)
{
	if (punch->end_page <= idx || punch->start_page >= idx + add)
		return;

	if (punch->start_page > idx) {
		ptr += punch->start_page - idx;
		add -= punch->start_page - idx;
		idx = punch->start_page;
	}
	if (punch->end_page < idx + add) {
		add = punch->end_page - idx;
	}

	punch->freed += nvfs_free_run(nmi->vfs_inode.i_sb->s_fs_info, ptr, add, punch->zero);
}

static void walk_indirect(struct nvfs_memory_inode *nmi, __le64 *ptr, unsigned level, uint64_t idx, struct punch_info *punch)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	unsigned long nodes_per_page = 1UL << (nvs->log2_page_size - 3);
	uint64_t add = 1ULL << (nvs->log2_page_size - 3) * level;
	__le64 *b;
	bool free_me = punch->start_page <= idx && punch->end_page >= idx + add * nodes_per_page;

	cond_resched();

	if (!le64_to_cpu(*ptr))
		return;
	b = (__le64 *)(nvs->memory_map + le64_to_cpu(*ptr));
	if (!level) {
		walk_direct(nmi, b, nodes_per_page, idx, punch);
	} else {
		unsigned long i;
					/* !!! TODO: this loop could be optimized */
		for (i = 0; i < nodes_per_page; i++) {
			if (punch->end_page <= idx || punch->start_page >= idx + add)
				goto skip_punch;
			walk_indirect(nmi, b + i, level - 1, idx, punch);

skip_punch:
			idx += add;
		}
	}

	if (free_me) {
		uint64_t p = le64_to_cpu(*ptr);
		if (punch->zero) {
			nv_store(*ptr, cpu_to_le64(0));
			wmb();
		}
		nvfs_free_blocks(nvs, p, 1);
		punch->freed++;
	}
}

int nvfs_punch_pages(struct nvfs_memory_inode *nmi, uint64_t start_page, uint64_t end_page, bool zero)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	struct nvfs_inode *nvi = nmi->nvi;
	unsigned long nodes_per_page = 1UL << (nvs->log2_page_size - 3);
	unsigned long npi;
	unsigned level;
	uint64_t idx;
	struct punch_info punch;
	bool quota;

	punch.start_page = start_page;
	punch.end_page = end_page;
	punch.zero = zero;
	punch.freed = 0;

	walk_direct(nmi, &nvi->i_direct_blocks[0], NVFS_DIRECT_BLOCKS, 0, &punch);

	idx = NVFS_DIRECT_BLOCKS;
	npi = nodes_per_page;

	for (level = 0; level < NVFS_INDIRECT_BLOCKS; level++) {
		walk_indirect(nmi, &nvi->i_indirect_blocks[level], level, idx, &punch);
		idx += npi;
		npi <<= nvs->log2_page_size - 3;
	}

	wmb();

	quota = unlikely((nvs->mount_flags & (NVFS_USR_QUOTA | NVFS_GRP_QUOTA)) != 0);
	if (unlikely(quota))
		dquot_free_space_nodirty(&nmi->vfs_inode, punch.freed << nvs->log2_page_size);
	nvfs_adj_block_count(nmi, -punch.freed, quota);

	return 0;
}

int nvfs_punch_hole(struct nvfs_memory_inode *nmi, loff_t start, loff_t end)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	uint64_t start_page, end_page;
	int r;
	unsigned start_offset, end_offset;

	if (unlikely(start < 0))
		start = 0;
	if (unlikely(end < 0))
		return 0;
	if (unlikely(start >= end))
		return 0;
	if (unlikely(end >= nmi->vfs_inode.i_size))
		end = LLONG_MAX - ((1UL << nvs->log2_page_size) - 1);

	down_write(&nmi->dax_sem);

	truncate_pagecache_range(&nmi->vfs_inode, start, end);

	start_page = (start + (1UL << nvs->log2_page_size) - 1) >> nvs->log2_page_size;
	end_page = end >> nvs->log2_page_size;

	start_offset = start & ((1UL << nvs->log2_page_size) - 1);
	end_offset = end & ((1UL << nvs->log2_page_size) - 1);

	if (start_page < end_page) {
		r = nvfs_punch_pages(nmi, start_page, end_page, true);
		if (unlikely(r < 0))
			goto unlock_ret_r;
	}

	if (start_page > end_page) {
		void *map = nvfs_bmap(nmi, end_page, NULL, NULL, NULL, NULL);
		if (map) {
			if (unlikely(IS_ERR(map))) {
				r = PTR_ERR(map);
				goto unlock_ret_r;
			}
			nvfs_memzero_flushcache(map + start_offset, end_offset - start_offset);
			cond_resched();
		}
	} else {
		if (start_offset) {
			void *map = nvfs_bmap(nmi, start_page - 1, NULL, NULL, NULL, NULL);
			if (map) {
				if (unlikely(IS_ERR(map))) {
					r = PTR_ERR(map);
					goto unlock_ret_r;
				}
				nvfs_memzero_flushcache(map + start_offset, (1UL << nvs->log2_page_size) - start_offset);
				cond_resched();
			}
		}
		if (end_offset) {
			void *map = nvfs_bmap(nmi, end_page, NULL, NULL, NULL, NULL);
			if (map) {
				if (unlikely(IS_ERR(map))) {
					r = PTR_ERR(map);
					goto unlock_ret_r;
				}
				nvfs_memzero_flushcache(map, end_offset);
				cond_resched();
			}
		}
	}

	r = 0;

unlock_ret_r:
	up_write(&nmi->dax_sem);

	return r;
}
